"""Titiler Enums."""

from enum import Enum
from types import DynamicClassAttribute

from rasterio.enums import Resampling

from rio_tiler.colormap import cmap
from rio_tiler.profiles import img_profiles


class MimeTypes(str, Enum):
    """MineTypes."""

    tif = "image/tiff; application=geotiff"
    jp2 = "image/jp2"
    png = "image/png"
    pngraw = "image/png"
    jpeg = "image/jpeg"
    webp = "image/webp"
    npy = "application/x-binary"
    pbf = "application/x-protobuf"
    mvt = "application/x-protobuf"
    xml = "application/xml"
    json = "application/json"
    html = "text/html"
    text = "text/plain"


class ImageDrivers(str, Enum):
    """Rio-tiler supported output drivers."""

    jpeg = "JPEG"
    png = "PNG"
    pngraw = "PNG"
    tif = "GTiff"
    webp = "WEBP"
    jp2 = "JP2OpenJPEG"
    npy = "NPY"


class ImageType(Enum):
    """Available Output Image type."""

    png = "png"
    npy = "npy"
    tif = "tif"
    jpeg = "jpg"
    jp2 = "jp2"
    webp = "webp"
    pngraw = "pngraw"

    @DynamicClassAttribute
    def profile(self):
        """Return rio-tiler image default profile."""
        return img_profiles.get(self._name_, {})

    @DynamicClassAttribute
    def driver(self):
        """Return rio-tiler image default profile."""
        return ImageDrivers[self._name_].value

    @DynamicClassAttribute
    def mimetype(self):
        """Return image mimetype."""
        return MimeTypes[self._name_].value


class TileType(Enum):
    """Available Output Tile type."""

    png = "png"
    npy = "npy"
    tif = "tif"
    jpeg = "jpg"
    jp2 = "jp2"
    webp = "webp"
    pngraw = "pngraw"
    pbf = "pbf"
    mvt = "mvt"

    @DynamicClassAttribute
    def profile(self):
        """Return rio-tiler image default profile."""
        return img_profiles.get(self._name_, {})

    @DynamicClassAttribute
    def driver(self):
        """Return rio-tiler image default profile."""
        try:
            return ImageDrivers[self._name_].value
        except KeyError:
            return ""

    @DynamicClassAttribute
    def mimetype(self):
        """Return image mimetype."""
        return MimeTypes[self._name_].value


class ColormapEnumFactory(Enum):
    """Rio-Tiler colormaps."""

    @DynamicClassAttribute
    def data(self):
        """Return rio-tiler image default profile."""
        return cmap.get(self._name_)


ColorMaps = ColormapEnumFactory(  # type: ignore
    "ColorMaps", [(a, a) for a in sorted(cmap.list())]
)

ResamplingNames = Enum(  # type: ignore
    "ResamplingNames", [(r.name, r.name) for r in Resampling]
)



"""Compatibility layer.

Create an AsyncBaseReader from a BaseReader subclass.

"""

from typing import Any, Coroutine, Dict, List, Tuple, Type

import attr
import morecantile
from starlette.concurrency import run_in_threadpool

from rio_tiler import constants
from rio_tiler.io import AsyncBaseReader, BaseReader, COGReader
from rio_tiler.models import ImageData, ImageStatistics, Info, Metadata


@attr.s
class AsyncReader(AsyncBaseReader):
    """Async Reader class."""

    src_path: str = attr.ib()
    tms: morecantile.TileMatrixSet = attr.ib(default=constants.WEB_MERCATOR_TMS)

    reader: Type[BaseReader] = COGReader

    def __attrs_post_init__(self):
        """PostInit."""
        self.dataset = self.reader(self.src_path)
        self.bounds = self.dataset.bounds
        self.minzoom = self.dataset.minzoom
        self.maxzoom = self.dataset.maxzoom

        self.assets = getattr(self.dataset, "assets", None)
        self.bands = getattr(self.dataset, "bands", None)
        self.colormap = getattr(self.dataset, "colormap", None)

        return self

    def close(self):
        """Close rasterio dataset."""
        if getattr(self.dataset, "close", None):
            self.dataset.close()

    async def __aexit__(self, exc_type, exc_value, traceback):
        """Support using with Context Managers."""
        self.close()

    async def info(self, **kwargs: Any) -> Coroutine[Any, Any, Info]:
        """Return Dataset's info."""
        return await run_in_threadpool(self.dataset.info, **kwargs)  # type: ignore

    async def stats(
        self, pmin: float = 2.0, pmax: float = 98.0, **kwargs: Any
    ) -> Coroutine[Any, Any, Dict[str, ImageStatistics]]:
        """Return Dataset's statistics."""
        return await run_in_threadpool(self.dataset.stats, pmin, pmax, **kwargs)  # type: ignore

    async def metadata(
        self, pmin: float = 2.0, pmax: float = 98.0, **kwargs: Any
    ) -> Coroutine[Any, Any, Metadata]:
        """Return Dataset's statistics."""
        return await run_in_threadpool(self.dataset.metadata, pmin, pmax, **kwargs)  # type: ignore

    async def tile(
        self, tile_x: int, tile_y: int, tile_z: int, **kwargs: Any
    ) -> Coroutine[Any, Any, ImageData]:
        """Read a Map tile from the Dataset."""
        return await run_in_threadpool(
            self.dataset.tile, tile_x, tile_y, tile_z, **kwargs  # type: ignore
        )

    async def point(
        self, lon: float, lat: float, **kwargs: Any
    ) -> Coroutine[Any, Any, List]:
        """Read a value from a Dataset."""
        return await run_in_threadpool(self.dataset.point, lon, lat, **kwargs)  # type: ignore

    async def part(
        self, bbox: Tuple[float, float, float, float], **kwargs: Any
    ) -> Coroutine[Any, Any, ImageData]:
        """Read a Part of a Dataset."""
        return await run_in_threadpool(self.dataset.part, bbox, **kwargs)  # type: ignore

    async def preview(self, **kwargs: Any) -> Coroutine[Any, Any, ImageData]:
        """Return a preview of a Dataset."""
        return await run_in_threadpool(self.dataset.preview, **kwargs)  # type: ignore

    async def feature(
        self, shape: Dict, **kwargs: Any
    ) -> Coroutine[Any, Any, ImageData]:
        """Return a preview of a Dataset."""
        return await run_in_threadpool(self.dataset.feature, **kwargs)  # type: ignore



"""rio-viz utility functions."""

import time


# This code is copied from marblecutter
#  https://github.com/mojodna/marblecutter/blob/master/marblecutter/stats.py
# License:
# Original work Copyright 2016 Stamen Design
# Modified work Copyright 2016-2017 Seth Fitzsimmons
# Modified work Copyright 2016 American Red Cross
# Modified work Copyright 2016-2017 Humanitarian OpenStreetMap Team
# Modified work Copyright 2017 Mapzen
class Timer(object):
    """Time a code block."""

    def __enter__(self):
        """Starts timer."""
        self.start = time.time()
        return self

    def __exit__(self, ty, val, tb):
        """Stops timer."""
        self.end = time.time()
        self.elapsed = self.end - self.start

    @property
    def from_start(self):
        """Return time elapsed from start."""
        return time.time() - self.start



"""rio-viz app dependencies."""

from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union

from fastapi import Query

from .resources.enums import ColorMaps, ResamplingNames


@dataclass
class ImageParams:
    """Image Rendering options."""

    rescale: Optional[str] = Query(
        None,
        title="Min/Max data Rescaling",
        description="comma (',') delimited Min,Max bounds",
    )
    color_formula: Optional[str] = Query(
        None,
        title="Color Formula",
        description="rio-color formula (info: https://github.com/mapbox/rio-color)",
    )
    color_map: Optional[ColorMaps] = Query(  # type: ignore
        None, description="rio-tiler's colormap name"
    )
    resampling_method: ResamplingNames = Query(
        None, description="Resampling method."  # type: ignore
    )
    rescale_range: Optional[List[Union[float, int]]] = field(init=False)
    colormap: Optional[Dict[int, Tuple[int, int, int, int]]] = field(init=False)

    def __post_init__(self):
        """Post Init."""
        if self.resampling_method is None:
            self.resampling_method = ResamplingNames.nearest

        self.rescale_range = (
            list(map(float, self.rescale.split(","))) if self.rescale else None
        )

        self.colormap = self.color_map.data if self.color_map else None


@dataclass
class DefaultDependency:
    """Dependency Base Class"""

    kwargs: Dict = field(init=False, default_factory=dict)


# Dependencies for simple BaseReader
@dataclass
class IndexesParams(DefaultDependency):
    """Band Indexes parameters."""

    bidx: Optional[List[int]] = Query(
        None, title="Band indexes", description="band indexes",
    )

    def __post_init__(self):
        """Post Init."""
        if self.bidx is not None:
            self.kwargs["indexes"] = self.bidx


# Dependencies for  MultiBaseReader
@dataclass
class AssetsParams(DefaultDependency):
    """Asset and Band indexes parameters."""

    asset: Optional[List[str]] = Query(
        None, title="Asset indexes", description="asset names."
    )
    bidx: Optional[List[int]] = Query(
        None, title="Band indexes", description="band indexes",
    )

    default_asset: Optional[List[str]] = field(init=False)

    def __post_init__(self):
        """Post Init."""
        if self.asset or self.default_asset:
            self.kwargs["assets"] = self.asset or self.default_asset
        if self.bidx is not None:
            self.kwargs["indexes"] = self.bidx


# Dependencies for  MultiBandReader
@dataclass
class BandsParams(DefaultDependency):
    """Band names parameters."""

    band: Optional[List[str]] = Query(
        None, title="bands names", description="bands names.",
    )
    bidx: Optional[List[int]] = Query(
        [1], title="Band indexes", description="band indexes",
    )

    default_band: Optional[List[str]] = field(init=False)

    def __post_init__(self):
        """Post Init."""
        if self.band or self.default_band:
            self.kwargs["bands"] = self.band or self.default_band
        if self.bidx is not None:
            self.kwargs["indexes"] = self.bidx



"""rio_viz app."""

import pathlib
import urllib.parse
from functools import partial
from typing import Any, Dict, Optional, Type, Union

import attr
import rasterio
import uvicorn
from fastapi import APIRouter, Depends, FastAPI, HTTPException, Query
from starlette.concurrency import run_in_threadpool
from starlette.middleware.cors import CORSMiddleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.requests import Request
from starlette.responses import HTMLResponse
from starlette.templating import Jinja2Templates

from rio_tiler.io import AsyncBaseReader
from rio_tiler.models import Info, Metadata

from .dependencies import (
    AssetsParams,
    BandsParams,
    DefaultDependency,
    ImageParams,
    IndexesParams,
)
from .models.mapbox import TileJSON
from .resources.enums import ImageType, TileType
from .resources.responses import ImageResponse, XMLResponse
from .utils import Timer

try:
    from rio_tiler_mvt.mvt import encoder as mvt_encoder  # noqa

    has_mvt = True
except ModuleNotFoundError:
    has_mvt = False
    mvt_encoder = None


template_dir = str(pathlib.Path(__file__).parent.joinpath("templates"))
templates = Jinja2Templates(directory=template_dir)


@attr.s
class viz:
    """Creates a very minimal slippy map tile server using fastAPI + Uvicorn."""

    src_path: str = attr.ib()
    reader: Type[AsyncBaseReader] = attr.ib()

    app: FastAPI = attr.ib(default=attr.Factory(FastAPI))

    token: Optional[str] = attr.ib(default=None)
    port: int = attr.ib(default=8080)
    host: str = attr.ib(default="127.0.0.1")
    style: str = attr.ib(default="satellite")
    config: Dict = attr.ib(default=dict)

    minzoom: Optional[int] = attr.ib(default=None)
    maxzoom: Optional[int] = attr.ib(default=None)

    layers: Optional[str] = attr.ib(default=None)
    nodata: Optional[Union[str, int, float]] = attr.ib(default=None)

    # cog / bands / assets
    reader_type: str = attr.ib(default="cog")

    router: Optional[APIRouter] = attr.ib(init=False)

    layer_dependency: Type[DefaultDependency] = attr.ib(init=False)

    @reader_type.validator
    def check(self, attribute, value):
        """Validate reader_type."""
        if value not in ["cog", "bands", "assets"]:
            raise ValueError("`reader_type` must be one of `cog, bands or assets`")

    def __attrs_post_init__(self):
        """Update App."""
        self.router = APIRouter()

        if self.reader_type == "cog":
            self.layer_dependency = IndexesParams
        elif self.reader_type == "bands":
            self.layer_dependency = type(
                "BandsParams",
                (BandsParams,),
                {"default_band": self.layers.split(",") if self.layers else None},
            )
        elif self.reader_type == "assets":
            self.layer_dependency = type(
                "AssetsParams",
                (AssetsParams,),
                {"default_asset": self.layers.split(",") if self.layers else None},
            )

        self.register_middleware()
        self.register_routes()
        self.app.include_router(self.router)

    def register_middleware(self):
        """Register Middleware to the FastAPI app."""
        self.app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["GET"],
            allow_headers=["*"],
        )
        self.app.add_middleware(GZipMiddleware, minimum_size=0)

    def _get_options(self, src_dst, options: Dict[str, Any]):
        """Create Reader options."""
        assets = getattr(src_dst, "assets", None)
        if assets and not options.get("assets"):
            options["assets"] = assets

        bands = getattr(src_dst, "bands", None)
        if bands and not options.get("bands"):
            options["bands"] = bands

        return options

    def register_routes(self):
        """Register routes to the FastAPI app."""

        @self.router.get(
            r"/tiles/{z}/{x}/{y}",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                        "application/x-protobuf": {},
                    },
                    "description": "Return a tile.",
                }
            },
            response_class=ImageResponse,
            description="Read COG and return a tile",
        )
        @self.router.get(
            r"/tiles/{z}/{x}/{y}.{format}",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                        "application/x-protobuf": {},
                    },
                    "description": "Return an image.",
                }
            },
            response_class=ImageResponse,
            description="Read COG and return a tile",
        )
        async def tile(
            z: int,
            x: int,
            y: int,
            scale: int = Query(2, gt=0, lt=4),
            format: Optional[TileType] = None,
            layer_params=Depends(self.layer_dependency),
            image_params: ImageParams = Depends(),
            feature_type: str = Query(
                None,
                title="Feature type (Only for Vector)",
                regex="^(point)|(polygon)$",
            ),
        ):
            """Handle /tiles requests."""
            timings = []
            headers: Dict[str, str] = {}

            tilesize = scale * 256

            if format and format in [TileType.pbf, TileType.mvt]:
                tilesize = 128

            with Timer() as t:
                async with self.reader(self.src_path) as src_dst:  # type: ignore

                    # Adapt options for each reader type
                    kwargs = self._get_options(src_dst, layer_params.kwargs)
                    if self.nodata is not None:
                        kwargs["nodata"] = self.nodata

                    tile_data = await src_dst.tile(
                        x,
                        y,
                        z,
                        tilesize=tilesize,
                        resampling_method=image_params.resampling_method.name,
                        **kwargs,
                    )

                    bandnames = kwargs.get("bands", kwargs.get("assets", None)) or [
                        f"{ix + 1}" for ix in range(tile_data.count)
                    ]

                    dst_colormap = getattr(src_dst, "colormap", None)
            timings.append(("dataread", round(t.elapsed * 1000, 2)))

            if format and format in [TileType.pbf, TileType.mvt]:
                if not mvt_encoder:
                    raise HTTPException(
                        status_code=500,
                        detail="rio-tiler-mvt not found, please do pip install rio-viz['mvt']",
                    )

                _mvt_encoder = partial(run_in_threadpool, mvt_encoder)

                with Timer() as t:
                    content = await _mvt_encoder(
                        tile_data.data,
                        tile_data.mask,
                        bandnames,
                        feature_type=feature_type,
                    )  # type: ignore
                timings.append(("format", round(t.elapsed * 1000, 2)))
            else:
                if not format:
                    format = TileType.jpeg if tile_data.mask.all() else TileType.png

                with Timer() as t:
                    image = tile_data.post_process(
                        in_range=image_params.rescale_range,
                        color_formula=image_params.color_formula,
                    )
                timings.append(("postprocess", round(t.elapsed * 1000, 2)))

                with Timer() as t:
                    content = image.render(
                        img_format=format.driver,
                        colormap=image_params.colormap or dst_colormap,
                        **format.profile,
                    )
                timings.append(("format", round(t.from_start * 1000, 2)))

            headers["Server-Timing"] = ", ".join(
                [f"{name};dur={time}" for (name, time) in timings]
            )
            return ImageResponse(content, media_type=format.mimetype, headers=headers)

        @self.router.get(
            r"/preview",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                    },
                    "description": "Return a preview.",
                }
            },
            response_class=ImageResponse,
            description="Return a preview",
        )
        @self.router.get(
            r"/preview.{format}",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                    },
                    "description": "Return a preview.",
                }
            },
            response_class=ImageResponse,
            description="Return a preview.",
        )
        async def preview(
            format: Optional[ImageType] = None,
            indexes: Optional[str] = Query(
                None, description="Coma (',') delimited band indexes"
            ),
            max_size: int = 1024,
            layer_params=Depends(self.layer_dependency),
            image_params: ImageParams = Depends(),
        ):
            """Handle /preview requests."""
            timings = []
            headers: Dict[str, str] = {}

            with Timer() as t:
                async with self.reader(self.src_path) as src_dst:  # type: ignore

                    # Adapt options for each reader type
                    kwargs = self._get_options(src_dst, layer_params.kwargs)
                    if self.nodata is not None:
                        kwargs["nodata"] = self.nodata

                    data = await src_dst.preview(
                        max_size=max_size,
                        resampling_method=image_params.resampling_method.name,
                        **kwargs,
                    )
                    dst_colormap = getattr(src_dst, "colormap", None)
            timings.append(("dataread", round(t.elapsed * 1000, 2)))

            if not format:
                format = ImageType.jpeg if data.mask.all() else ImageType.png

            with Timer() as t:
                image = data.post_process(
                    in_range=image_params.rescale_range,
                    color_formula=image_params.color_formula,
                )
            timings.append(("postprocess", round(t.elapsed * 1000, 2)))

            with Timer() as t:
                content = image.render(
                    img_format=format.driver,
                    colormap=image_params.colormap or dst_colormap,
                    **format.profile,
                )
            timings.append(("format", round(t.from_start * 1000, 2)))

            headers["Server-Timing"] = ", ".join(
                [f"{name};dur={time}" for (name, time) in timings]
            )
            return ImageResponse(content, media_type=format.mimetype, headers=headers)

        @self.router.get(
            r"/part",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                    },
                    "description": "Return a part of a dataset.",
                }
            },
            response_class=ImageResponse,
            description="Return a preview.",
        )
        @self.router.get(
            r"/part.{format}",
            responses={
                200: {
                    "content": {
                        "image/png": {},
                        "image/jpeg": {},
                        "image/webp": {},
                        "image/jp2": {},
                        "image/tiff; application=geotiff": {},
                        "application/x-binary": {},
                    },
                    "description": "Return a part of a dataset.",
                }
            },
            response_class=ImageResponse,
            description="Return a preview.",
        )
        async def part(
            format: Optional[ImageType] = Query(None, description="Output image type."),
            bbox: str = Query(
                ..., description="Bounding box in form of 'minx,miny,maxx,maxy'"
            ),
            indexes: Optional[str] = Query(
                None, description="Coma (',') delimited band indexes"
            ),
            max_size: int = 1024,
            layer_params=Depends(self.layer_dependency),
            image_params: ImageParams = Depends(),
        ):
            """Handle /part requests."""
            timings = []
            headers: Dict[str, str] = {}

            with Timer() as t:
                async with self.reader(self.src_path) as src_dst:  # type: ignore

                    # Adapt options for each reader type
                    kwargs = self._get_options(src_dst, layer_params.kwargs)
                    if self.nodata is not None:
                        kwargs["nodata"] = self.nodata

                    data = await src_dst.part(
                        list(map(float, bbox.split(","))),
                        max_size=max_size,
                        resampling_method=image_params.resampling_method.name,
                        **kwargs,
                    )
                    dst_colormap = getattr(src_dst, "colormap", None)
            timings.append(("dataread", round(t.elapsed * 1000, 2)))

            if not format:
                format = ImageType.jpeg if data.mask.all() else ImageType.png

            with Timer() as t:
                image = data.post_process(
                    in_range=image_params.rescale_range,
                    color_formula=image_params.color_formula,
                )
            timings.append(("postprocess", round(t.elapsed * 1000, 2)))

            with Timer() as t:
                content = image.render(
                    img_format=format.driver,
                    colormap=image_params.colormap or dst_colormap,
                    **format.profile,
                )
            timings.append(("format", round(t.from_start * 1000, 2)))

            headers["Server-Timing"] = ", ".join(
                [f"{name};dur={time}" for (name, time) in timings]
            )
            return ImageResponse(content, media_type=format.mimetype, headers=headers)

        @self.router.get(
            "/point", responses={200: {"description": "Return a point value."}},
        )
        async def point(
            coordinates: str = Query(
                ..., description="Coma (',') delimited lon,lat coordinates"
            ),
            layer_params=Depends(self.layer_dependency),
        ):
            """Handle /point requests."""
            lon, lat = list(map(float, coordinates.split(",")))
            async with self.reader(self.src_path) as src_dst:  # type: ignore
                kwargs = self._get_options(src_dst, layer_params.kwargs)
                if self.nodata is not None:
                    kwargs["nodata"] = self.nodata

                results = await src_dst.point(lon, lat, **kwargs)

            return {"coordinates": [lon, lat], "value": results}

        @self.router.get(
            "/metadata",
            response_model=Union[Dict[str, Metadata], Metadata],
            response_model_exclude={"minzoom", "maxzoom", "center"},
            response_model_exclude_none=True,
            responses={200: {"description": "Return the metadata of the COG."}},
        )
        async def metadata(
            pmin: float = 2.0,
            pmax: float = 98.0,
            max_size: int = 1024,
            layer_params=Depends(self.layer_dependency),
        ):
            """Handle /metadata requests."""
            async with self.reader(self.src_path) as src_dst:  # type: ignore
                # Adapt options for each reader type
                kwargs = self._get_options(src_dst, layer_params.kwargs)
                if self.nodata is not None:
                    kwargs["nodata"] = self.nodata

                return await src_dst.metadata(pmin, pmax, max_size=max_size, **kwargs)

        @self.router.get(
            "/info",
            response_model=Union[Dict[str, Info], Info],
            response_model_exclude={"minzoom", "maxzoom", "center"},
            response_model_exclude_none=True,
            responses={200: {"description": "Return the metadata of the COG."}},
        )
        async def info(layer_params=Depends(self.layer_dependency)):
            """Handle /info requests."""
            async with self.reader(self.src_path) as src_dst:  # type: ignore
                # Adapt options for each reader type
                kwargs = self._get_options(src_dst, layer_params.kwargs)
                return await src_dst.info(**kwargs)

        @self.router.get(
            "/tilejson.json",
            response_model=TileJSON,
            responses={200: {"description": "Return a tilejson"}},
            response_model_exclude_none=True,
        )
        async def tilejson(
            request: Request,
            tile_format: Optional[TileType] = None,
            layer_params=Depends(self.layer_dependency),  # noqa
            image_params: ImageParams = Depends(),  # noqa
            feature_type: str = Query(  # noqa
                None, title="Feature type", regex="^(point)|(polygon)$"
            ),
        ):
            """Handle /tilejson.json requests."""
            kwargs: Dict[str, Any] = {"z": "{z}", "x": "{x}", "y": "{y}"}
            if tile_format:
                kwargs["format"] = tile_format.value

            tile_url = request.url_for("tile", **kwargs)

            qs = [
                (key, value)
                for (key, value) in request.query_params._list
                if key != "tile_format"
            ]
            if qs:
                tile_url += f"?{urllib.parse.urlencode(qs)}"

            async with self.reader(self.src_path) as src_dst:  # type: ignore
                bounds = src_dst.bounds
                center = src_dst.center
                minzoom = self.minzoom if self.minzoom is not None else src_dst.minzoom
                maxzoom = self.maxzoom if self.maxzoom is not None else src_dst.maxzoom

            return dict(
                bounds=bounds,
                center=center,
                minzoom=minzoom,
                maxzoom=maxzoom,
                name="rio-viz",
                tilejson="2.1.0",
                tiles=[tile_url],
            )

        @self.router.get(
            "/index.html",
            responses={200: {"description": "Simple COG viewer."}},
            response_class=HTMLResponse,
        )
        async def viewer(request: Request):
            """Handle /index.html."""
            if self.reader_type == "cog":
                name = "index.html"
            elif self.reader_type == "bands":
                name = "bands.html"
            elif self.reader_type == "assets":
                name = "assets.html"

            return templates.TemplateResponse(
                name=name,
                context={
                    "request": request,
                    "tilejson_endpoint": request.url_for("tilejson"),
                    "metadata_endpoint": request.url_for("metadata"),
                    "point_endpoint": request.url_for("point"),
                    "mapbox_access_token": self.token,
                    "mapbox_style": self.style,
                    "allow_3d": has_mvt,
                },
                media_type="text/html",
            )

        @self.router.get("/WMTSCapabilities.xml", response_class=XMLResponse)
        async def wmts(
            request: Request,
            tile_format: ImageType = Query(
                ImageType.png, description="Output image type. Default is png."
            ),
            layer_params=Depends(self.layer_dependency),  # noqa
            image_params: ImageParams = Depends(),  # noqa
            feature_type: str = Query(  # noqa
                None, title="Feature type", regex="^(point)|(polygon)$"
            ),
        ):
            """
            This is a hidden gem.

            rio-viz is meant to be use to visualize your dataset in the browser but
            using this endpoint, you can also load it in you GIS software.

            """
            kwargs = {
                "z": "{TileMatrix}",
                "x": "{TileCol}",
                "y": "{TileRow}",
                "format": tile_format.value,
            }
            tiles_endpoint = request.url_for("tile", **kwargs)

            q = dict(request.query_params)
            q.pop("tile_format", None)
            q.pop("minzoom", None)
            q.pop("maxzoom", None)
            q.pop("SERVICE", None)
            q.pop("REQUEST", None)
            qs = urllib.parse.urlencode(list(q.items()))
            if qs:
                tiles_endpoint += f"?{qs}"

            async with self.reader(self.src_path) as src_dst:  # type: ignore
                bounds = src_dst.bounds
                minzoom = self.minzoom if self.minzoom is not None else src_dst.minzoom
                maxzoom = self.maxzoom if self.maxzoom is not None else src_dst.maxzoom

            tileMatrix = []
            for zoom in range(minzoom, maxzoom + 1):
                tm = f"""<TileMatrix>
                    <ows:Identifier>{zoom}</ows:Identifier>
                    <ScaleDenominator>{559082264.02872 / 2 ** zoom / 1}</ScaleDenominator>
                    <TopLeftCorner>-20037508.34278925 20037508.34278925</TopLeftCorner>
                    <TileWidth>256</TileWidth>
                    <TileHeight>256</TileHeight>
                    <MatrixWidth>{2 ** zoom}</MatrixWidth>
                    <MatrixHeight>{2 ** zoom}</MatrixHeight>
                </TileMatrix>"""
                tileMatrix.append(tm)

            return templates.TemplateResponse(
                "wmts.xml",
                {
                    "request": request,
                    "tiles_endpoint": tiles_endpoint,
                    "bounds": bounds,
                    "tileMatrix": tileMatrix,
                    "title": "Cloud Optimized GeoTIFF",
                    "layer_name": "cogeo",
                    "media_type": tile_format.mimetype,
                },
                media_type="application/xml",
            )

    @property
    def endpoint(self) -> str:
        """Get endpoint url."""
        return f"http://{self.host}:{self.port}"

    @property
    def template_url(self) -> str:
        """Get simple app template url."""
        return f"http://{self.host}:{self.port}/index.html"

    @property
    def docs_url(self) -> str:
        """Get simple app template url."""
        return f"http://{self.host}:{self.port}/docs"

    def start(self):
        """Start tile server."""
        with rasterio.Env(**self.config):
            uvicorn.run(app=self.app, host=self.host, port=self.port, log_level="info")



"""rio-viz multifile reader."""

from typing import Dict, Type

import attr
from braceexpand import braceexpand
from morecantile import TileMatrixSet

from rio_tiler.constants import WEB_MERCATOR_TMS
from rio_tiler.errors import InvalidBandName
from rio_tiler.io import BaseReader, COGReader, MultiBandReader


@attr.s
class MultiFilesReader(MultiBandReader):
    """Multiple Files reader.

    Args:
        filepath (str): Brace Expandable path (e.g: band{1,2,3}.tif).

    """

    filepath: str = attr.ib()
    reader: Type[BaseReader] = attr.ib(default=COGReader)
    reader_options: Dict = attr.ib(factory=dict)
    tms: TileMatrixSet = attr.ib(default=WEB_MERCATOR_TMS)

    def __attrs_post_init__(self):
        """Fetch Reference band to get the bounds."""
        self.files = list(braceexpand(self.filepath))
        self.indexes = [ix + 1 for ix in range(len(self.files))]
        self.bands = [f"file{ix}" for ix in self.indexes]

        with self.reader(self.files[0], tms=self.tms, **self.reader_options) as cog:
            self.bounds = cog.bounds
            self.minzoom = cog.minzoom
            self.maxzoom = cog.maxzoom

    def _get_band_url(self, band: str) -> str:
        """Validate band's name and return band's url."""
        if band not in self.bands:
            raise InvalidBandName(f"{band} is not valid")

        index = self.bands.index(band)
        return self.files[index]



"""rio_viz.cli."""

import importlib
import os
import tempfile
import warnings
from contextlib import ExitStack, contextmanager

import click
import numpy
from rasterio.rio import options
from rio_cogeo.cogeo import cog_translate, cog_validate
from rio_cogeo.profiles import cog_profiles

from rio_tiler.io import (
    AsyncBaseReader,
    BaseReader,
    COGReader,
    MultiBandReader,
    MultiBaseReader,
)
from rio_viz import app
from rio_viz.compat import AsyncReader


@contextmanager
def TemporaryRasterFile(suffix=".tif"):
    """Create temporary file."""
    fileobj = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
    fileobj.close()
    try:
        yield fileobj
    finally:
        os.remove(fileobj.name)


class MbxTokenType(click.ParamType):
    """Mapbox token type."""

    name = "token"

    def convert(self, value, param, ctx):
        """Validate token."""
        try:
            if not value:
                return ""

            assert value.startswith("pk")
            return value

        except (AttributeError, AssertionError):
            raise click.ClickException(
                "Mapbox access token must be public (pk). "
                "Please sign up at https://www.mapbox.com/signup/ to get a public token. "
                "If you already have an account, you can retreive your "
                "token at https://www.mapbox.com/account/."
            )


class NodataParamType(click.ParamType):
    """Nodata inddex type."""

    name = "nodata"

    def convert(self, value, param, ctx):
        """Validate and parse band index."""
        try:
            if value.lower() == "nan":
                return numpy.nan
            elif value.lower() in ["nil", "none", "nada"]:
                return None
            else:
                return float(value)
        except (TypeError, ValueError):
            raise click.ClickException("{} is not a valid nodata value.".format(value))


@click.command()
@click.argument("src_path", type=str, nargs=1, required=True)
@click.option(
    "--nodata",
    type=NodataParamType(),
    metavar="NUMBER|nan",
    help="Set nodata masking values for input dataset.",
)
@click.option(
    "--minzoom", type=int, help="Overwrite minzoom",
)
@click.option(
    "--maxzoom", type=int, help="Overwrite maxzoom",
)
@click.option(
    "--style",
    type=click.Choice(["dark", "satellite", "basic"]),
    default="dark",
    help="Mapbox basemap",
)
@click.option("--port", type=int, default=8080, help="Webserver port (default: 8080)")
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Webserver host url (default: 127.0.0.1)",
)
@click.option(
    "--mapbox-token",
    type=MbxTokenType(),
    metavar="TOKEN",
    default=lambda: os.environ.get("MAPBOX_ACCESS_TOKEN", ""),
    help="Pass Mapbox token",
)
@click.option("--no-check", is_flag=True, help="Ignore COG validation")
@click.option(
    "--reader",
    type=str,
    help="rio-tiler Reader (BaseReader or AsyncBaseReader). Default is `rio_tiler.io.COGReader`",
)
@click.option(
    "--layers",
    type=str,
    help="limit to specific layers (only used for MultiBand and MultiBase Readers).",
)
@click.option(
    "--server-only",
    is_flag=True,
    default=False,
    help="Launch API without opening the rio-viz web-page.",
)
@click.option(
    "--config",
    "config",
    metavar="NAME=VALUE",
    multiple=True,
    callback=options._cb_key_val,
    help="GDAL configuration options.",
)
def viz(
    src_path,
    nodata,
    minzoom,
    maxzoom,
    style,
    port,
    host,
    mapbox_token,
    no_check,
    reader,
    layers,
    server_only,
    config,
):
    """Rasterio Viz cli."""

    if reader:
        module, classname = reader.rsplit(".", 1)
        reader = getattr(importlib.import_module(module), classname)  # noqa
        if not issubclass(reader, (BaseReader, AsyncBaseReader)):
            warnings.warn("Reader should be a subclass of rio_tiler.io.BaseReader")

    dataset_reader = reader or COGReader

    if issubclass(dataset_reader, (MultiBandReader)):
        reader_type = "bands"
    elif issubclass(dataset_reader, (MultiBaseReader)):
        reader_type = "assets"
    else:
        reader_type = "cog"

    # Check if cog
    with ExitStack() as ctx:
        if (
            src_path.lower().endswith(".tif")
            and not reader
            and not no_check
            and not cog_validate(src_path)[0]
        ):
            # create tmp COG
            click.echo("create temporaty COG")
            tmp_path = ctx.enter_context(TemporaryRasterFile())

            output_profile = cog_profiles.get("deflate")
            output_profile.update(dict(blockxsize="256", blockysize="256"))
            config = dict(GDAL_TIFF_INTERNAL_MASK=True, GDAL_TIFF_OVR_BLOCKSIZE="128")
            cog_translate(src_path, tmp_path.name, output_profile, config=config)
            src_path = tmp_path.name

        # Dynamically create an Async Dataset Reader if not a subclass of AsyncBaseReader
        if not issubclass(dataset_reader, AsyncBaseReader):
            dataset_reader = type(
                "AsyncReader", (AsyncReader,), {"reader": dataset_reader}
            )

        application = app.viz(
            src_path=src_path,
            reader=dataset_reader,
            token=mapbox_token,
            port=port,
            host=host,
            style=style,
            config=config,
            minzoom=minzoom,
            maxzoom=maxzoom,
            nodata=nodata,
            layers=layers,
            reader_type=reader_type,
        )
        if not server_only:
            click.echo(f"Viewer started at {application.template_url}", err=True)
            click.launch(application.template_url)

        application.start()



"""setup: rio-viz."""

from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

# Runtime requirements.
inst_reqs = [
    "braceexpand",
    "click",
    "fastapi==0.63.0",
    "jinja2 >=2.11.2,<3.0.0",
    "rasterio",
    "rio-cogeo>=2.1,<2.2",
    "rio-tiler>=2.0,<2.1",
    "uvicorn[standard]>=0.12.0,<0.14.0",
    # Additional requirements for python 3.6
    "dataclasses;python_version<'3.7'",
    "async_exit_stack>=1.0.1,<2.0.0;python_version<'3.7'",
    "async_generator>=1.10,<2.0.0;python_version<'3.7'",
]

extra_reqs = {
    "mvt": ["rio-tiler-mvt"],
    "test": ["pytest", "pytest-cov", "pytest-asyncio", "requests"],
    "dev": ["pytest", "pytest-cov", "pytest-asyncio", "pre-commit"],
}

setup(
    name="rio-viz",
    version="0.6.0",
    python_requires=">=3.6",
    description=u"Visualize Cloud Optimized GeoTIFF in browser",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: GIS",
    ],
    keywords="COG COGEO Rasterio GIS MVT",
    author=u"Vincent Sarago",
    author_email="vincent@developmentseed.org",
    url="https://github.com/developmentseed/rio-viz",
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    zip_safe=False,
    license="MIT",
    install_requires=inst_reqs,
    extras_require=extra_reqs,
    entry_points="""
      [rasterio.rio_plugins]
      viz=rio_viz.scripts.cli:viz
      """,
)
